@@ -1,4 +1,5 @@
 import datetime
+import ipaddress
 import logging
 import os
 import ssl
@@ -48,6 +49,9 @@ TLS_VERSION_1_3_DRAFT_28 = 0x7F1C
 TLS_VERSION_1_3_DRAFT_27 = 0x7F1B
 TLS_VERSION_1_3_DRAFT_26 = 0x7F1A
 
+CLIENT_CONTEXT_STRING = b"TLS 1.3, client CertificateVerify"
+SERVER_CONTEXT_STRING = b"TLS 1.3, server CertificateVerify"
+
 T = TypeVar("T")
 
 # facilitate mocking for the test suite
@@ -137,14 +141,16 @@ class State(Enum):
     CLIENT_EXPECT_SERVER_HELLO = 1
     CLIENT_EXPECT_ENCRYPTED_EXTENSIONS = 2
     CLIENT_EXPECT_CERTIFICATE_REQUEST_OR_CERTIFICATE = 3
-    CLIENT_EXPECT_CERTIFICATE_CERTIFICATE = 4
+    CLIENT_EXPECT_CERTIFICATE = 4
     CLIENT_EXPECT_CERTIFICATE_VERIFY = 5
     CLIENT_EXPECT_FINISHED = 6
     CLIENT_POST_HANDSHAKE = 7
 
     SERVER_EXPECT_CLIENT_HELLO = 8
-    SERVER_EXPECT_FINISHED = 9
-    SERVER_POST_HANDSHAKE = 10
+    SERVER_EXPECT_CERTIFICATE = 9
+    SERVER_EXPECT_CERTIFICATE_VERIFY = 10
+    SERVER_EXPECT_FINISHED = 11
+    SERVER_POST_HANDSHAKE = 12
 
 
 def hkdf_label(label: bytes, hash_value: bytes, length: int) -> bytes:
@@ -218,13 +224,36 @@ def verify_certificate(
     # verify subject
     if server_name is not None:
         try:
-            service_identity.cryptography.verify_certificate_hostname(
-                certificate, server_name
-            )
+            ipaddress.ip_address(server_name)
+        except ValueError:
+            is_ip = False
+        else:
+            is_ip = True
+
+        try:
+            if is_ip:
+                service_identity.cryptography.verify_certificate_ip_address(
+                    certificate, server_name
+                )
+            else:
+                service_identity.cryptography.verify_certificate_hostname(
+                    certificate, server_name
+                )
+
         except service_identity.VerificationError as exc:
-            raise AlertBadCertificate(
-                "Certificate does not match hostname '%s'" % server_name
-            ) from exc
+            patterns = service_identity.cryptography.extract_patterns(certificate)
+            if len(patterns) == 0:
+                errmsg = "subject alternative name not found in the certificate"
+            elif len(patterns) == 1:
+                errmsg = f"hostname {server_name!r} doesn't match {patterns[0]!r}"
+            else:
+                patterns_repr = ", ".join(repr(pattern) for pattern in patterns)
+                errmsg = (
+                    f"hostname {server_name!r} doesn't match "
+                    f"either of {patterns_repr}"
+                )
+
+            raise AlertBadCertificate(errmsg) from exc
 
     # load CAs
     store = crypto.X509Store()
@@ -837,6 +866,55 @@ def push_certificate(buf: Buffer, certificate: Certificate) -> None:
         )
 
 
+@dataclass
+class CertificateRequest:
+    request_context: bytes = b""
+    signature_algorithms: Optional[List[int]] = None
+    other_extensions: List[Tuple[int, bytes]] = field(default_factory=list)
+
+
+def pull_certificate_request(buf: Buffer) -> CertificateRequest:
+    certificate_request = CertificateRequest()
+
+    assert buf.pull_uint8() == HandshakeType.CERTIFICATE_REQUEST
+    with pull_block(buf, 3):
+        certificate_request.request_context = pull_opaque(buf, 1)
+
+        def pull_extension() -> None:
+            extension_type = buf.pull_uint16()
+            extension_length = buf.pull_uint16()
+            if extension_type == ExtensionType.SIGNATURE_ALGORITHMS:
+                certificate_request.signature_algorithms = pull_list(
+                    buf, 2, buf.pull_uint16
+                )
+            else:
+                certificate_request.other_extensions.append(
+                    (extension_type, buf.pull_bytes(extension_length))
+                )
+
+        pull_list(buf, 2, pull_extension)
+
+    return certificate_request
+
+
+def push_certificate_request(
+    buf: Buffer, certificate_request: CertificateRequest
+) -> None:
+    buf.push_uint8(HandshakeType.CERTIFICATE_REQUEST)
+    with push_block(buf, 3):
+        push_opaque(buf, 1, certificate_request.request_context)
+
+        with push_block(buf, 2):
+            with push_extension(buf, ExtensionType.SIGNATURE_ALGORITHMS):
+                push_list(
+                    buf, 2, buf.push_uint16, certificate_request.signature_algorithms
+                )
+
+            for extension_type, extension_value in certificate_request.other_extensions:
+                with push_extension(buf, extension_type):
+                    buf.push_bytes(extension_value)
+
+
 @dataclass
 class CertificateVerify:
     algorithm: int
@@ -1114,8 +1192,10 @@ class Context:
             Union[dsa.DSAPrivateKey, ec.EllipticCurvePrivateKey, rsa.RSAPrivateKey]
         ] = None
         self.handshake_extensions: List[Extension] = []
+        self._is_client = is_client
         self._max_early_data = max_early_data
         self.session_ticket: Optional[SessionTicket] = None
+        self._request_client_certificate = False  # For test purposes only
         self._server_name = server_name
         if verify_mode is not None:
             self._verify_mode = verify_mode
@@ -1163,11 +1243,13 @@ class Context:
         self.early_data_accepted = False
         self.key_schedule: Optional[KeySchedule] = None
         self.received_extensions: Optional[List[Extension]] = None
+        self._certificate_request: Optional[CertificateRequest] = None
         self._key_schedule_psk: Optional[KeySchedule] = None
         self._key_schedule_proxy: Optional[KeyScheduleProxy] = None
         self._new_session_ticket: Optional[NewSessionTicket] = None
         self._peer_certificate: Optional[x509.Certificate] = None
         self._peer_certificate_chain: List[x509.Certificate] = []
+        self._psk_key_exchange_mode: Optional[int] = None
         self._receive_buffer = b""
         self._session_resumed = False
         self._enc_key: Optional[bytes] = None
@@ -1230,10 +1312,16 @@ class Context:
                 else:
                     raise AlertUnexpectedMessage
             elif self.state == State.CLIENT_EXPECT_CERTIFICATE_REQUEST_OR_CERTIFICATE:
+                if message_type == HandshakeType.CERTIFICATE:
+                    self._client_handle_certificate(input_buf)
+                elif message_type == HandshakeType.CERTIFICATE_REQUEST:
+                    self._client_handle_certificate_request(input_buf)
+                else:
+                    raise AlertUnexpectedMessage
+            elif self.state == State.CLIENT_EXPECT_CERTIFICATE:
                 if message_type == HandshakeType.CERTIFICATE:
                     self._client_handle_certificate(input_buf)
                 else:
-                    # FIXME: handle certificate request
                     raise AlertUnexpectedMessage
             elif self.state == State.CLIENT_EXPECT_CERTIFICATE_VERIFY:
                 if message_type == HandshakeType.CERTIFICATE_VERIFY:
@@ -1263,6 +1351,20 @@ class Context:
                     )
                 else:
                     raise AlertUnexpectedMessage
+            elif self.state == State.SERVER_EXPECT_CERTIFICATE:
+                if message_type == HandshakeType.CERTIFICATE:
+                    self._server_handle_certificate(
+                        input_buf, output_buf[Epoch.ONE_RTT]
+                    )
+                else:
+                    raise AlertUnexpectedMessage
+            elif self.state == State.SERVER_EXPECT_CERTIFICATE_VERIFY:
+                if message_type == HandshakeType.CERTIFICATE_VERIFY:
+                    self._server_handle_certificate_verify(
+                        input_buf, output_buf[Epoch.ONE_RTT]
+                    )
+                else:
+                    raise AlertUnexpectedMessage
             elif self.state == State.SERVER_EXPECT_FINISHED:
                 if message_type == HandshakeType.FINISHED:
                     self._server_handle_finished(input_buf, output_buf[Epoch.ONE_RTT])
@@ -1299,6 +1401,20 @@ class Context:
             ticket=new_session_ticket.ticket,
         )
 
+    def _check_certificate_verify_signature(self, verify: CertificateVerify) -> None:
+        assert verify.algorithm in self._signature_algorithms
+
+        try:
+            self._peer_certificate.public_key().verify(
+                verify.signature,
+                self.key_schedule.certificate_verify_data(
+                    SERVER_CONTEXT_STRING if self._is_client else CLIENT_CONTEXT_STRING
+                ),
+                *signature_algorithm_params(verify.algorithm),
+            )
+        except InvalidSignature:
+            raise AlertDecryptError
+
     def _client_send_hello(self, output_buf: Buffer) -> None:
         key_share: List[KeyShareEntry] = []
         supported_groups: List[int] = []
@@ -1326,6 +1442,15 @@ class Context:
 
         assert len(key_share), "no key share entries"
 
+        # Literal IPv4 and IPv6 addresses are not permitted in
+        # Server Name Indication (SNI) hostname.
+        try:
+            ipaddress.ip_address(self._server_name)
+        except ValueError:
+            server_name = self._server_name
+        else:
+            server_name = None
+
         hello = ClientHello(
             random=self.client_random,
             legacy_session_id=self.legacy_session_id,
@@ -1336,7 +1461,7 @@ class Context:
             psk_key_exchange_modes=self._psk_key_exchange_modes
             if (self.session_ticket or self.new_session_ticket_cb is not None)
             else None,
-            server_name=self._server_name,
+            server_name=server_name,
             signature_algorithms=self._signature_algorithms,
             supported_groups=supported_groups,
             supported_versions=self._supported_versions,
@@ -1468,37 +1593,23 @@ class Context:
         else:
             self._set_state(State.CLIENT_EXPECT_CERTIFICATE_REQUEST_OR_CERTIFICATE)
 
+    def _client_handle_certificate_request(self, input_buf: Buffer) -> None:
+        self._certificate_request = pull_certificate_request(input_buf)
+        self.key_schedule.update_hash(input_buf.data)
+        self._set_state(State.CLIENT_EXPECT_CERTIFICATE)
+
     def _client_handle_certificate(self, input_buf: Buffer) -> None:
         certificate = pull_certificate(input_buf)
-
-        self._peer_certificate = x509.load_der_x509_certificate(
-            certificate.certificates[0][0]
-        )
-        self._peer_certificate_chain = [
-            x509.load_der_x509_certificate(certificate.certificates[i][0])
-            for i in range(1, len(certificate.certificates))
-        ]
-
         self.key_schedule.update_hash(input_buf.data)
 
+        self._set_peer_certificate(certificate)
         self._set_state(State.CLIENT_EXPECT_CERTIFICATE_VERIFY)
 
     def _client_handle_certificate_verify(self, input_buf: Buffer) -> None:
         verify = pull_certificate_verify(input_buf)
 
-        assert verify.algorithm in self._signature_algorithms
-
         # check signature
-        try:
-            self._peer_certificate.public_key().verify(
-                verify.signature,
-                self.key_schedule.certificate_verify_data(
-                    b"TLS 1.3, server CertificateVerify"
-                ),
-                *signature_algorithm_params(verify.algorithm),
-            )
-        except InvalidSignature:
-            raise AlertDecryptError
+        self._check_certificate_verify_signature(verify)
 
         # check certificate
         if self._verify_mode != ssl.CERT_NONE:
@@ -1512,7 +1623,6 @@ class Context:
             )
 
         self.key_schedule.update_hash(input_buf.data)
-
         self._set_state(State.CLIENT_EXPECT_FINISHED)
 
     def _client_handle_finished(self, input_buf: Buffer, output_buf: Buffer) -> None:
@@ -1532,6 +1642,48 @@ class Context:
         )
         next_enc_key = self.key_schedule.derive_secret(b"c ap traffic")
 
+        if self._certificate_request is not None:
+            # check whether we have a suitable signature algorithm
+            if (
+                self.certificate is not None
+                and self.certificate_private_key is not None
+            ):
+                signature_algorithm = negotiate(
+                    self._signature_algorithms_for_private_key(),
+                    self._certificate_request.signature_algorithms,
+                )
+            else:
+                signature_algorithm = None
+
+            # send certificate
+            with push_message(self.key_schedule, output_buf):
+                push_certificate(
+                    output_buf,
+                    Certificate(
+                        request_context=self._certificate_request.request_context,
+                        certificates=[
+                            (x.public_bytes(Encoding.DER), b"")
+                            for x in [self.certificate] + self.certificate_chain
+                        ]
+                        if signature_algorithm
+                        else [],
+                    ),
+                )
+
+            # send certificate verify
+            if signature_algorithm:
+                signature = self.certificate_private_key.sign(
+                    self.key_schedule.certificate_verify_data(CLIENT_CONTEXT_STRING),
+                    *signature_algorithm_params(signature_algorithm),
+                )
+                with push_message(self.key_schedule, output_buf):
+                    push_certificate_verify(
+                        output_buf,
+                        CertificateVerify(
+                            algorithm=signature_algorithm, signature=signature
+                        ),
+                    )
+
         # send finished
         with push_message(self.key_schedule, output_buf):
             push_finished(
@@ -1562,6 +1714,39 @@ class Context:
             )
             self.new_session_ticket_cb(ticket)
 
+    def _server_expect_finished(self, onertt_buf: Buffer):
+        # anticipate client's FINISHED
+        self._expected_verify_data = self.key_schedule.finished_verify_data(
+            self._dec_key
+        )
+        buf = Buffer(capacity=64)
+        push_finished(buf, Finished(verify_data=self._expected_verify_data))
+        self.key_schedule.update_hash(buf.data)
+
+        # create a new session ticket
+        if (
+            self.new_session_ticket_cb is not None
+            and self._psk_key_exchange_mode is not None
+        ):
+            self._new_session_ticket = NewSessionTicket(
+                ticket_lifetime=86400,
+                ticket_age_add=struct.unpack("I", os.urandom(4))[0],
+                ticket_nonce=b"",
+                ticket=os.urandom(64),
+                max_early_data_size=self._max_early_data,
+            )
+
+            # send message
+            push_new_session_ticket(onertt_buf, self._new_session_ticket)
+
+            # notify application
+            ticket = self._build_session_ticket(
+                self._new_session_ticket, self.handshake_extensions
+            )
+            self.new_session_ticket_cb(ticket)
+
+        self._set_state(State.SERVER_EXPECT_FINISHED)
+
     def _server_handle_hello(
         self,
         input_buf: Buffer,
@@ -1571,23 +1756,6 @@ class Context:
     ) -> None:
         peer_hello = pull_client_hello(input_buf)
 
-        # determine applicable signature algorithms
-        signature_algorithms: List[SignatureAlgorithm] = []
-        if isinstance(self.certificate_private_key, rsa.RSAPrivateKey):
-            signature_algorithms = [
-                SignatureAlgorithm.RSA_PSS_RSAE_SHA256,
-                SignatureAlgorithm.RSA_PKCS1_SHA256,
-                SignatureAlgorithm.RSA_PKCS1_SHA1,
-            ]
-        elif isinstance(
-            self.certificate_private_key, ec.EllipticCurvePrivateKey
-        ) and isinstance(self.certificate_private_key.curve, ec.SECP256R1):
-            signature_algorithms = [SignatureAlgorithm.ECDSA_SECP256R1_SHA256]
-        elif isinstance(self.certificate_private_key, ed25519.Ed25519PrivateKey):
-            signature_algorithms = [SignatureAlgorithm.ED25519]
-        elif isinstance(self.certificate_private_key, ed448.Ed448PrivateKey):
-            signature_algorithms = [SignatureAlgorithm.ED448]
-
         # negotiate parameters
         cipher_suite = negotiate(
             self._cipher_suites,
@@ -1603,7 +1771,7 @@ class Context:
             self._psk_key_exchange_modes, peer_hello.psk_key_exchange_modes
         )
         signature_algorithm = negotiate(
-            signature_algorithms,
+            self._signature_algorithms_for_private_key(),
             peer_hello.signature_algorithms,
             AlertHandshakeFailure("No supported signature algorithm"),
         )
@@ -1747,6 +1915,17 @@ class Context:
             )
 
         if pre_shared_key is None:
+            # send certificate request
+            if self._request_client_certificate:
+                with push_message(self.key_schedule, handshake_buf):
+                    push_certificate_request(
+                        handshake_buf,
+                        CertificateRequest(
+                            request_context=b"",
+                            signature_algorithms=self._signature_algorithms,
+                        ),
+                    )
+
             # send certificate
             with push_message(self.key_schedule, handshake_buf):
                 push_certificate(
@@ -1762,9 +1941,7 @@ class Context:
 
             # send certificate verify
             signature = self.certificate_private_key.sign(
-                self.key_schedule.certificate_verify_data(
-                    b"TLS 1.3, server CertificateVerify"
-                ),
+                self.key_schedule.certificate_verify_data(SERVER_CONTEXT_STRING),
                 *signature_algorithm_params(signature_algorithm),
             )
             with push_message(self.key_schedule, handshake_buf):
@@ -1792,34 +1969,32 @@ class Context:
         )
         self._next_dec_key = self.key_schedule.derive_secret(b"c ap traffic")
 
-        # anticipate client's FINISHED as we don't use client auth
-        self._expected_verify_data = self.key_schedule.finished_verify_data(
-            self._dec_key
-        )
-        buf = Buffer(capacity=64)
-        push_finished(buf, Finished(verify_data=self._expected_verify_data))
-        self.key_schedule.update_hash(buf.data)
+        self._psk_key_exchange_mode = psk_key_exchange_mode
+        if self._request_client_certificate:
+            self._set_state(State.SERVER_EXPECT_CERTIFICATE)
+        else:
+            self._server_expect_finished(onertt_buf)
 
-        # create a new session ticket
-        if self.new_session_ticket_cb is not None and psk_key_exchange_mode is not None:
-            self._new_session_ticket = NewSessionTicket(
-                ticket_lifetime=86400,
-                ticket_age_add=struct.unpack("I", os.urandom(4))[0],
-                ticket_nonce=b"",
-                ticket=os.urandom(64),
-                max_early_data_size=self._max_early_data,
-            )
+    def _server_handle_certificate(self, input_buf: Buffer, output_buf: Buffer) -> None:
+        certificate = pull_certificate(input_buf)
+        self.key_schedule.update_hash(input_buf.data)
 
-            # send message
-            push_new_session_ticket(onertt_buf, self._new_session_ticket)
+        if certificate.certificates:
+            self._set_peer_certificate(certificate)
+            self._set_state(State.SERVER_EXPECT_CERTIFICATE_VERIFY)
+        else:
+            self._server_expect_finished(output_buf)
 
-            # notify application
-            ticket = self._build_session_ticket(
-                self._new_session_ticket, self.handshake_extensions
-            )
-            self.new_session_ticket_cb(ticket)
+    def _server_handle_certificate_verify(
+        self, input_buf: Buffer, output_buf: Buffer
+    ) -> None:
+        verify = pull_certificate_verify(input_buf)
 
-        self._set_state(State.SERVER_EXPECT_FINISHED)
+        # check signature
+        self._check_certificate_verify_signature(verify)
+
+        self.key_schedule.update_hash(input_buf.data)
+        self._server_expect_finished(output_buf)
 
     def _server_handle_finished(self, input_buf: Buffer, output_buf: Buffer) -> None:
         finished = pull_finished(input_buf)
@@ -1854,7 +2029,34 @@ class Context:
             direction, epoch, self.key_schedule.cipher_suite, key
         )
 
+    def _set_peer_certificate(self, certificate: Certificate) -> None:
+        self._peer_certificate = x509.load_der_x509_certificate(
+            certificate.certificates[0][0]
+        )
+        self._peer_certificate_chain = [
+            x509.load_der_x509_certificate(certificate.certificates[i][0])
+            for i in range(1, len(certificate.certificates))
+        ]
+
     def _set_state(self, state: State) -> None:
         if self.__logger:
             self.__logger.debug("TLS %s -> %s", self.state, state)
         self.state = state
+
+    def _signature_algorithms_for_private_key(self) -> List[SignatureAlgorithm]:
+        signature_algorithms: List[SignatureAlgorithm] = []
+        if isinstance(self.certificate_private_key, rsa.RSAPrivateKey):
+            signature_algorithms = [
+                SignatureAlgorithm.RSA_PSS_RSAE_SHA256,
+                SignatureAlgorithm.RSA_PKCS1_SHA256,
+                SignatureAlgorithm.RSA_PKCS1_SHA1,
+            ]
+        elif isinstance(
+            self.certificate_private_key, ec.EllipticCurvePrivateKey
+        ) and isinstance(self.certificate_private_key.curve, ec.SECP256R1):
+            signature_algorithms = [SignatureAlgorithm.ECDSA_SECP256R1_SHA256]
+        elif isinstance(self.certificate_private_key, ed25519.Ed25519PrivateKey):
+            signature_algorithms = [SignatureAlgorithm.ED25519]
+        elif isinstance(self.certificate_private_key, ed448.Ed448PrivateKey):
+            signature_algorithms = [SignatureAlgorithm.ED448]
+        return signature_algorithms
